{% extends "!breadcrumbs.html" %}
{% block breadcrumbs_aside %}
<li class="wy-breadcrumbs-aside">
    {% set pth = pathto('.') %}
    {% if 'hoc/' in pagename %}
        {% set switch = '../' + pagename | replace("hoc/", "python/") + '.html' %}
        <a href="{{ pth }}/{{ switch }}" > {{ _('Switch to Python') }}</a>
    {% elif 'python/' in pagename %}
        {% set switch = '../' + pagename | replace("python/", "hoc/") + '.html' %}
        <a href="{{ pth }}/{{ switch }}" >  {{ _('Switch to HOC') }}</a>
    {% endif %}
</li>
{{ super() }}
{% endblock %}
