

#define yymaxdepth difeq_yymaxdepth
#define yyparse    difeq_yyparse
#define yylex      difeq_yylex
#define yyerror    difeq_yyerror
#define yylval     difeq_yylval
#define yychar     difeq_yychar
#define yydebug    difeq_yydebug
#define yypact     difeq_yypact
#define yyr1       difeq_yyr1
#define yyr2       difeq_yyr2
#define yydef      difeq_yydef
#define yychk      difeq_yychk
#define yypgo      difeq_yypgo
#define yyact      difeq_yyact
#define yyexca     difeq_yyexca
#define yyerrflag  difeq_yyerrflag
#define yynerrs    difeq_yynerrs
#define yyps       difeq_yyps
#define yypv       difeq_yypv
#define yys        difeq_yys
#define yy_yys     difeq_yyyys
#define yystate    difeq_yystate
#define yytmp      difeq_yytmp
#define yyv        difeq_yyv
#define yy_yyv     difeq_yyyyv
#define yyval      difeq_yyval
#define yylloc     difeq_yylloc
#define yyreds     difeq_yyreds
#define yytoks     difeq_yytoks
#define yylhs      difeq_yyyylhs
#define yylen      difeq_yyyylen
#define yydefred   difeq_yyyydefred
#define yydgoto    difeq_yyyydgoto
#define yysindex   difeq_yyyysindex
#define yyrindex   difeq_yyyyrindex
#define yygindex   difeq_yyyygindex
#define yytable    difeq_yyyytable
#define yycheck    difeq_yyyycheck
