"at_time", "", "ms", 0, "state_discontinuity", "", "m9", "m9", 0, "net_send", "", "ms", "m9", 0,
    "net_event", "", "ms", 0, "net_move", "", "ms", 0, "nrn_ghk", "millicoulombs/cm3", "millivolt",
    "milli/liter", "milli/liter", "", 0,
