#ifndef bbssrv_h
#define bbssrv_h

#define QUIT             0
#define POST             1
#define LOOK             2
#define LOOK_TAKE        3
#define TAKE             4
#define LOOK_YES         6
#define LOOK_NO          7
#define LOOK_TAKE_YES    8
#define LOOK_TAKE_NO     9
#define HELLO            10
#define POST_TODO        11
#define POST_RESULT      12
#define LOOK_TAKE_TODO   13
#define LOOK_TAKE_RESULT 14
#define TAKE_TODO        15
#define CONTEXT          16
#define CRAY_POST        17
#define CRAY_POST_TODO   18
#define CRAY_POST_RESULT 19
#define FIRSTID          20
#endif
