/* auto include file for metrowerks codewarrior for all nrn */
#if __MWERKS__ >= 7
#define _MSL_DIRENT_H
#else
#include <Win32Headers.mch>
#endif
#define HAVE_LIMITS_H 1
#pragma once off
#ifndef __WIN32__
#define __WIN32__ 1
#endif
#ifndef WIN32
#define WIN32 1
#endif
#define _WIN32
#ifndef _Windows
#define _Windows 1
#endif
#if __cplusplus
struct Section;
struct Object;
struct Symbol;
#endif
#define motif_kit
#define sgi_motif_kit
#define printf  myprintf
#define vprintf myvprintf
#define gets    mygets
#define puts    myputs
#define fprintf myfprintf
#undef small
#undef near
#define small mysmall
#define near  mynear

#define stricmp _stricmp
#define putenv  _putenv

//#define system mysystem
#undef DELETE
#undef IGNORE
#define CABLE      1
#define HOC        1
#define OOP        1
#define OC_CLASSES "nrnclass.h"
#define USECVODE   1
#define CVODE      1
#define USEMATRIX  1
