/* /local/src/master/nrn/src/nmodl/extdef.h,v 4.2 1998/07/04 14:08:32 hines Exp */

"first_time", "error", "f_flux", "b_flux", "fabs", "sqrt", "sin", "cos", "tan", "acos", "asin",
    "atan", "atan2", "sinh", "cosh", "tanh", "floor", "ceil", "fmod", "log10", "log", "pow",
    "printf", "prterr", "exp", "threshold", "force", "deflate", "expfit", "derivs", "spline",
    "hyperbol", "revhyperbol", "sigmoid", "revsigmoid", "harmonic", "squarewave", "sawtooth",
    "revsawtooth", "ramp", "pulse", "perpulse", "step", "perstep", "erf", "exprand", "factorial",
    "gauss", "normrand", "poisrand", "poisson", "setseed", "scop_random", "boundary", "romberg",
    "legendre", "invert", "stepforce", "schedule", "set_seed", "nrn_pointing",
    "state_discontinuity", "net_send", "net_move", "net_event", "nrn_random_play", "at_time",
    "nrn_ghk",
