/*
 * Automake doesn't deal well with source files in a separate directory, so
 * this is just a stub to trick it into working right.
 */

#define OC_CLASSES "nrnclass.h"
#include "../ivoc/classreg.cpp"
