/* /local/src/master/nrn/src/oc/redef.h,v 1.3 1994/10/28 15:18:01 hines Exp */

#ifndef hoc_redef_h
#define hoc_redef_h

#define Break              hoc_Break
#define Continue           hoc_Continue
#define Exp                hoc_Exp
#define Fprint             hoc_Fprint
#define Fscan              hoc_Fscan
#define Log                hoc_Log
#define Log10              hoc_Log10
#define Plt                hoc_Plt
#define plt                hoc_plt
#define color              hoc_color
#define Setcolor           hoc_Setcolor
#define set_color          hoc_set_color
#define Pow                hoc_Pow
#define PRintf             hoc_PRintf
#define Sqrt               hoc_Sqrt
#define System             hoc_System
#define Xred               hoc_Xred
#define acterror           hoc_acterror
#define add                hoc_add
#define arayinstal         hoc_arayinstal
#define free_arrayinfo     hoc_free_arrayinfo
#define araypt             hoc_araypt
#define arg                hoc_arg
#define argassign          hoc_argassign
#define assign             hoc_assign
#define assstr             hoc_assstr
#define begin              hoc_begin
#define bltin              hoc_bltin
#define call               hoc_call
#define call_ob_proc       hoc_call_ob_proc
#define cbuf               hoc_cbuf
#define Code               hoc_Code
#define codei              hoc_codei
#define codesym            hoc_codesym
#define codein             hoc_codein
#define constpush          hoc_constpush
#define pushzero           hoc_pushzero
#define ctp                hoc_ctp
#define debug              hoc_debug
#define debugzz            hoc_debugzz
#define defnonly           hoc_defnonly
#define dep_make           hoc_dep_make
#define do_equation        hoc_do_equation
#define eq                 hoc_eq
#define eqinit             hoc_eqinit
#define eqn_init           hoc_eqn_init
#define eqn_lhs            hoc_eqn_lhs
#define eqn_name           hoc_eqn_name
#define eqn_rhs            hoc_eqn_rhs
#define eqn_side           hoc_eqn_side
#define errcheck           hoc_errcheck
#define eval               hoc_eval
#define execerror          hoc_execerror
#define execute            hoc_execute
#define fin                hoc_fin
#define follow             hoc_follow
#define forcode            hoc_forcode
#define shortfor           hoc_shortfor
#define fout               hoc_fout
#define fpecatch           hoc_fpecatch
#define free_list          hoc_free_list
#define frin               hoc_frin
#define funcret            hoc_funcret
#define gargstr            hoc_gargstr
#define getarg             hoc_getarg
#define gargc              hoc_gargc
#define gargv              hoc_gargv
#define ge                 hoc_ge
#define getnb              hoc_getnb
#define gt                 hoc_gt
#define ifcode             hoc_ifcode
#define indef              hoc_indef
#define infile             hoc_infile
#define init_access        hoc_init_access
#define initcode           hoc_initcode
#define insertcode         hoc_insertcode
#define install            hoc_install
#define integer            hoc_integer
#define intset             hoc_intset
#define ired               hoc_ired
#define xred               hoc_xred
#define sred               hoc_sred
#define le                 hoc_le
#define lineno             hoc_lineno
#define lookup             hoc_lookup
#define lt                 hoc_lt
#define moreinput          hoc_moreinput
#define mul                hoc_mul
#define ne                 hoc_ne
#define negate             hoc_negate
#define onintr             hoc_onintr
#define p_symlist          hoc_p_symlist
#define parserror          hoc_parserror
#define pc                 hoc_pc
#define pipeflag           hoc_pipeflag
#define pipegetc           hoc_pipegetc
#define pipeungetc         hoc_pipeungetc
#define xpop               hoc_xpop
#define spop               hoc_spop
#define ipop               hoc_ipop
#define nopop              hoc_nopop
#define power              hoc_power
#define prexpr             hoc_prexpr
#define print              hoc_print
#define plprint            hoc_plprint
#define procret            hoc_procret
#define prog               hoc_prog
#define prog_parse_recover hoc_prog_parse_recover
#define progbase           hoc_progbase
#define progname           hoc_progname
#define progp              hoc_progp
#define prstr              hoc_prstr
#define pushx              hoc_pushx
#define pushs              hoc_pushs
#define pushi              hoc_pushi
#define push_frame         hoc_push_frame
#define pop_frame          hoc_pop_frame
#define ret                hoc_ret
#define ropen              hoc_ropen
#define run                hoc_run
#define solve              hoc_solve
#define spinit             hoc_spinit
#define symbols            hoc_symbols
#define symlist            hoc_symlist
#define var_access         hoc_var_access
#define varpush            hoc_varpush
#define varread            hoc_varread
#define warning            hoc_warning
#define wopen              hoc_wopen
#define xopen              hoc_xopen
#define zzdebug            hoc_zzdebug

#endif
