#define u_char _lib_os(u_char)
#define CopyString _lib_os(CopyString)
#define List _lib_os(List)
#define Math _lib_os(Math)
#define NullTerminatedString _lib_os(NullTerminatedString)
#define PtrList _lib_os(PtrList)
#define String _lib_os(String)
