#pragma precompile_target "modlunit_def.h"
#define MSL_USE_PRECOMPILED_HEADERS 1
#include <ansi_prefix.mac.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#pragma once off
#define SYSV 1
#define MAC  1

#define NRNUNIT 1
