#ifndef nmodl2_redef_h
#define nmodl2_redef_h

#undef v
#undef area
#undef thisnode
#undef GC
#undef EC
#undef extnode
#undef xain
#undef xbout
#undef i
#undef sec

#undef Memb_list
#undef nodelist
#undef nodeindices
#undef data
#undef pdata
#undef prop
#undef nodecount
#undef pval

#undef type
#undef id

#endif
