"force", "deflate", "expfit", "derivs", "spline", "exprand", "gauss", "normrand", "poisrand",
    "poisson", "setseed", "scop_random", "boundary", "romberg", "invert", "stepforce", "schedule",
    "set_seed", "nrn_random_play",
